#region usings
using System;
using System.ComponentModel.Composition;

//using System.Drawing;

using SlimDX;
using SlimDX.Direct3D9;
using VVVV.PluginInterfaces.V1;
using VVVV.PluginInterfaces.V2;
using VVVV.PluginInterfaces.V2.EX9;
using VVVV.Utils.VColor;
using VVVV.Utils.VMath;
using VVVV.Utils.SlimDX;

using VVVV.Core.Logging;
#endregion usings

//here you can change the vertex type
using VertexType = VVVV.Utils.SlimDX.TexturedVertex;

namespace VVVV.Nodes
{
	#region PluginInfo
	[PluginInfo(Name = "Floyd", Category = "EX9.Texture", Version = "Simple", Help = "Basic template with one color in/out", Tags = "")]
	#endregion PluginInfo
	public class EX9TextureFloyd : DXTextureOutPluginBase, IPluginEvaluate
	{
		#region fields & pins
		[Input("Input")]
		ISpread<RGBAColor> FColorIn;
		
		[InputAttribute("Quantize", DefaultValue = 64, MinValue = 1, IsSingle = true)]
		ISpread<int> FStepIn;

		[Input("Width", DefaultValue = 64, IsSingle = true)]
		IDiffSpread<int> FWidthIn;

		[Input("Height", DefaultValue = 64, IsSingle = true)]
		IDiffSpread<int> FHeightIn;

		//[Output("Output")]
		//ISpread<RGBAColor> FColorOut;

		[Import()]
		ILogger FLogger;
		#endregion fields & pins

		int FCurrentSlice;
		
		public double[,] ditherArray;
		public double[,] lumaArray;

		// import host and hand it to base constructor
		[ImportingConstructor()]
		public EX9TextureFloyd(IPluginHost host) 
		    : base(host)
		{
		}
		
		
		//called when data for any output pin is requested
		public void Evaluate(int SpreadMax)
		{
			
			
			//recreate texture if resolution was changed
			if (FWidthIn.IsChanged || FHeightIn.IsChanged) 
			{
				SetSliceCount(1);
				Reinitialize();
			}
			
			int h = FHeightIn[0];
			int w = FWidthIn[0];
			ditherArray = new double[h, w];
			lumaArray = new double[h, w];
			for (int col = 0; col < w; col++) {
				for (int row = 0; row < h; row++) {
	 				var pxl = FColorIn[row * w + col];
					lumaArray[row,col] = GrauBerechnen(pxl.R, pxl.G, pxl.B);
				}
 			}
			Update();
	
			//FLogger.Log(LogType.Debug, "Logging to Renderer (TTY)");
		}
		
		public double GrauBerechnen(double rot, double gruen, double blau)
		{

				 return rot*0.299 + gruen*0.587 + blau*0.114;
		}
		
		public double DoSomething(double input)
		{
			return Math.Floor(input*FStepIn[0])/FStepIn[0];
		}

		protected override Texture CreateTexture(int Slice, Device device)
		{
			FLogger.Log(LogType.Debug, "Creating new texture at slice: " + Slice);
			return TextureUtils.CreateTexture(device, Math.Max(FWidthIn[Slice], 1), Math.Max(FHeightIn[Slice], 1));
		}
		

		protected unsafe override void UpdateTexture(int Slice, Texture texture)
		{
			FCurrentSlice = Slice;
			TextureUtils.Fill32BitTexInPlace(texture, FillTexure);
		}
		
		//this is a pixelshader like method, which we pass to the fill function
		private unsafe void FillTexure(uint* data, int row, int col, int width, int height)
		{
			var x = (double)row / height;
			var y = (double)col / width;
			
			ditherArray[row, col] =  lumaArray[row,col];
			lumaArray[row,col]=DoSomething(lumaArray[row,col]);
			double quant_error = ditherArray[row,col] - lumaArray[row,col];
	
			if (row +1 < height) 						lumaArray[row + 1,	col] += 7/16 * quant_error;
			if (row > 1 && col +1 < width)				lumaArray[row - 1 ,col + 1] += 3/16 * quant_error;
			if (col +1 < width) 						lumaArray[row  ,	col + 1] +=  5/16 * quant_error;
			if (row + 1 < height  && col +1 < width) 	lumaArray[row + 1 ,col + 1] += 1/16 * quant_error;	

			var wave = (byte)(255*(lumaArray[row,col]));
			var rrr = (byte)(col%255);
			var ggg = (byte)(row%255);
			var pixel = UInt32Utils.fromARGB(255, wave, wave, wave);
			TextureUtils.SetPtrVal2D(data, pixel, row, col, width);
		}
	}
}
